#!/bin/ksh
#
# pkgmake - make a Solaris SVR4 package
#
###
#
# Copyright (c) 2003-2018 Lighterra - Open Source (MPL & LGPL)
#
# THIS IS OPEN SOURCE CODE DUAL-LICENSED UNDER THE MPL AND THE LGPL.
# The basic intent is that:
#   1. You may combine either open or closed source code with this code,
#      and the result may be sold commercially if you wish.
#   2. If you make any changes or additions to this source code you must
#      publish them, typically by sending them back to the original author
#      or project to be integrated into the main public distribution.
#   3. You must acknowledge that your software uses this code by displaying
#      a suitable notice somewhere, typically in an "about" dialog box,
#      command-line version flag output, manual or other documentation.
# If you are unsure whether this licensing fits your particular situation,
# please contact the author or copyright holder for clarification.
#
# DUAL-LICENSING DETAILS
# This software is subject to the terms of the Mozilla Public License
# Version 2.0 (the "MPL"). You may obtain a copy of the MPL from
# www.mozilla.org/MPL. Alternatively, this software may be used under the
# terms of the GNU Lesser General Public License Version 2.1 (the "LGPL"),
# in which case the provisions of the LGPL are applicable instead of those
# of the MPL. You may obtain a copy of the LGPL from www.gnu.org/licenses.
# If you wish to allow use of your version of this software only under the
# terms of the LGPL, and not to allow others to use your version of this
# software under the MPL, indicate your decision by deleting the provisions
# above and replacing them with the notice and other provisions required by
# the LGPL. If you do not delete the provisions above, a recipient may use
# your version of this software under either the MPL or the LGPL.
#
# NO WARRANTY
# This software is distributed in the hope that it will be useful, however
# software distributed under either the MPL or the LGPL is distributed
# on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either expressed or
# implied; without even the implied warranty of MERCHANTABILITY OR FITNESS
# FOR A PARTICULAR PURPOSE. See the MPL and the LGPL for details.
#
###
#
# written by Jason Robert Carey Patterson <jason@lighterra.com>
#

export PATH="$PATH:/usr/bin:/usr/sbin"
set -e
umask 022

version="1.8 (Dec 2018)"


#
# parse command-line args
#
usage() { echo "usage: pkgmake [options] filesdir [destdir]" 1>&2; exit 1; }

filesdir=""
destdir=.
use_long_name=false
change_owners=true
match_sys_dirs=true
be_verbose=false

while getopts "l(long-name)o(owners)v(verbose)V(version)" opt; do
    case $opt in
        l) use_long_name=true ;;
        o) change_owners=false; match_sys_dirs=false ;;
        v) be_verbose=true ;;
        V) echo "pkgmake version $version"; usage ;;
        *) usage ;;
    esac
done
if [ $OPTIND -gt 1 ]; then shift $(($OPTIND - 1)); fi

if [ $# -lt 1 -o $# -gt 2 ]; then usage; fi

filesdir="$1"
if [ $# -ge 2 ]; then destdir="$2"; fi

for dir in "$filesdir" "$destdir"; do
    if [ ! -d "$dir" ]; then
        echo "pkgmake: $dir: No such directory" 1>&2
        exit 2
    fi
done

destdir="`(cd \"$destdir\"; echo \"$PWD\")`"    # need full path later


#
# error functions
#
verbose() { if $be_verbose; then echo "$*"; fi; }
warning() { echo "pkgmake: warning: $*" 1>&2; }
error() { echo "pkgmake: error: $*" 1>&2; exit 3; }


#
# copy files into temp working dir
#
verbose "copying files..."

if [ -x "`which mktemp`" ]; then
    tempdir="`mktemp -d -t pkgmake.XXXXXX`"
else
    tempdir="${TMPDIR:-/tmp}/pkgmake.$$"
    mkdir -m 700 "$tempdir"
fi
trap "cd /; rm -rf '$tempdir'" 0 1 2 3 15

cd "$filesdir"
mkdir "$tempdir/reloc"
tar -cf - . | (cd "$tempdir/reloc"; tar -xf -)


#
# check for special package files
#
verbose "checking for special package files..."

pkg_files="pkginfo.txt depend.txt compver.txt space.txt copyright.txt"
pkg_files="$pkg_files request.sh checkinstall.sh preinstall.sh postinstall.sh"
pkg_files="$pkg_files preremove.sh postremove.sh"

lowercase() { echo "$*" | tr '[A-Z]' '[a-z]'; }
uppercase() { echo "$*" | tr '[a-z]' '[A-Z]'; }

find_file() {    # check with & without extension, lower & upper case
    ext=`echo $1 | sed -e 's/.*\.//'`
    name=`basename $1 .$ext`
    for e in $ext `uppercase $ext`; do
        for n in $name `uppercase $name`; do
            for file in $n $n.$e; do
                if [ -f $file ]; then echo $file; return; fi
            done
        done
    done
}

cd "$tempdir/reloc"
for file in $pkg_files; do
    file=`find_file $file`
    if [ -n "$file" ]; then
        verbose "found $file"
        chmod 644 $file
        official_filename=`lowercase $file | sed -e 's/\..*//'`
        mv $file ../$official_filename
        echo "i $official_filename" >>../prototype-extras
    fi
done

if [ ! -f ../pkginfo ]; then
    error "a pkginfo file must be present in filesdir"
fi


#
# generate prototype file
#
file=`find_file prototype.txt`
if [ -n "$file" ]; then
    verbose "found $file"
    chmod 644 $file
    mv $file ../prototype
else
    verbose "generating prototype file..."
    find . | pkgproto >../prototype
    cat ../prototype-extras >>../prototype
fi

cd ..
rm prototype-extras


#
# examine pkginfo file
#
verbose "examining pkginfo file..."

pkginfo_field() { awk -F= "/^$1=/ { print \$2 }" <pkginfo | tr -d '"'; }

pkginfo_field_must_be_defined() {
    val="`pkginfo_field $1`"
    if [ -z "$val" ]; then error "pkginfo must contain an entry for $1"; fi
    echo "$val"
}

pkginfo_field_defaults_to() {
    val="`pkginfo_field $1`"
    if [ -z "$val" ]; then val="$2"; echo "$1=\"$val\"" >>pkginfo; fi
    echo "$val"
}

pkg="`pkginfo_field_must_be_defined PKG`"
if [ `echo "$pkg" | grep -c '^[A-Z][A-Z0-9]*[a-z][a-z0-9-+]*$'` -ne 1 ]; then
    error "pkginfo PKG field must be of the form COMPANYpackagename"
fi
if [ `echo "$pkg" | wc -c` -gt 33 ]; then
    error "pkginfo PKG field is too long (maximum 32 chars)"
fi

name="`pkginfo_field_must_be_defined NAME`"
version="`pkginfo_field_must_be_defined VERSION`"

arch="`pkginfo_field_must_be_defined ARCH`"
case "$arch" in
    native)                                               arch=`uname -p` ;;
    x86* | amd64* | i[4-9]86*) arch_for_filename="$arch"; arch=i386       ;;
    sparc*)                    arch_for_filename="$arch"; arch=sparc      ;;
esac
sed -e "s/^ARCH=.*/ARCH=\"$arch\"/" <pkginfo >pkginfo.tmp
mv pkginfo.tmp pkginfo
arch_for_filename="${arch_for_filename:-$arch}"
if [ "$arch_for_filename" = i386 ]; then arch_for_filename=x86; fi

basedir="`pkginfo_field_defaults_to BASEDIR /`"
category="`pkginfo_field_defaults_to CATEGORY application`"
classes="`pkginfo_field_defaults_to CLASSES none`"
maxinst="`pkginfo_field_defaults_to MAXINST 999999`"

date_time=`TZ=UTC date +%Y%m%d%H%M%S`
pstamp="`pkginfo_field_defaults_to PSTAMP $date_time`"


#
# examine depend file
#
if [ -f depend ]; then
    verbose "examining depend file..."
    sed -e "s/(native)/(`uname -p`)/g" <depend >depend.tmp
    mv depend.tmp depend
fi


#
# generate compver file
#
if [ ! -f compver ]; then
    verbose "generating compver file..."

    versions="`echo \"$version\" | tr '.' ' '`"
    maj=`echo "$versions" | awk '{ print $1 }' | sed -e 's/[^0-9].*//'`
    min=`echo "$versions" | awk '{ print $2 }' | sed -e 's/[^0-9].*//'`
    pat=`echo "$versions" | awk '{ print $3 }' | sed -e 's/[^0-9].*//'`
    case $maj in [0-9] | [0-9][0-9]) ;; *) maj=0 ;; esac
    case $min in [0-9] | [0-9][0-9]) ;; *) min=0 ;; esac
    case $pat in [0-9] | [0-9][0-9]) ;; *) pat=0 ;; esac

    # write out every possible version from 0.0.0 to $version
    # this loop nest might iterate up to a million times (typically ~30,000)
    if [ -x "`which perl`" ]; then    # Perl is way faster if available
        perl -e \
        "open(\$out, \">\", \"compver\")
             or die \"cannot create file: \$\!\\n\";
         for (\$i = 0; \$i <= $maj; \$i++) {
             print \$out \"\$i\\n\";
             for (\$j = 0; \$i < $maj && \$j <= 99 || \$j <= $min; \$j++) {
                 print \$out \"\$i.\$j\\n\";
                 for (\$k = 0; \$i < $maj && \$j <= 99 && \$k <= 99 ||
                               \$j < $min && \$k <= 99 || \$k <= $pat; \$k++) {
                     print \$out \"\$i.\$j.\$k\\n\";
                 }
             }
         }"
    else
        i=0
        while [ $i -le $maj ]; do
            echo $i >>compver
            j=0
            while [ $i -lt $maj -a $j -le 99 -o $j -le $min ]; do
                echo $i.$j >>compver
                k=0
                while [ $i -lt $maj -a $j -le 99 -a $k -le 99 -o \
                        $j -lt $min -a $k -le 99 -o $k -le $pat ]; do
                    echo $i.$j.$k >>compver
                    k=$(($k + 1))
                done
                j=$(($j + 1))
            done
            i=$(($i + 1))
        done
    fi

    echo "i compver" >>prototype
fi


#
# set ownership in prototype file
#
if $change_owners; then
    verbose "setting installed file ownership to root..."
    sed -e 's/ [^ ]* [^ ]*$/ root root/' <prototype >prototype.tmp
    mv prototype.tmp prototype
fi

if [ $match_sys_dirs = true -a "$basedir" = "/" ]; then
    verbose "matching ownership for system dirs..."
    grep '^d none ' <prototype | while read d none dir mode rest; do
        if [ -d "/$dir" ]; then
            ls_output="`ls -ld \"/$dir\"`"
            u=`echo $ls_output | awk '{ print $3 }'`
            g=`echo $ls_output | awk '{ print $4 }'`
            m=$mode

            case "/$dir" in  # use ? to leave /usr, /var, /etc & /opt unchanged
                /usr | /var | /var/log | /var/cache | /etc | /opt )
                    u=?; g=?; m=? ;;
            esac

            sed -e "s!^d none $dir [^ ]* [^ ]* [^ ]*\$!d none $dir $m $u $g!" \
                <prototype >prototype.tmp
            mv prototype.tmp prototype
        fi
    done
fi


#
# make the package
#
verbose "making the package..."

arch="$arch_for_filename"
if $use_long_name; then
    filename="$pkg-$version-solaris-$arch.pkg"
else
    filename="`echo $pkg | sed -e 's/^[A-Z0-9]*//'`-$version-solaris-$arch.pkg"
fi

if $be_verbose; then
    pkgmk -b "$tempdir/reloc" -d .
    pkgtrans -s . "$filename" "$pkg"
else
    set +e    # we want to see any error output we capture below

    pkgmk -b "$tempdir/reloc" -d . >output.txt 2>&1
    if [ $? != 0 ]; then cat output.txt 1>&2; exit 4; fi

    pkgtrans -s . "$filename" "$pkg" >output.txt 2>&1
    if [ $? != 0 ]; then cat output.txt 1>&2; exit 5; fi
fi

cp "$filename" "$destdir/$filename"
